import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class ink_well extends PApplet {

//Ink Well
//Kevin Schlei, 2017

//================================================================
//MARK: Properties
//================================================================

PVector bodyStartAnchorPoint, bodyEndAnchorPoint;
float kCameraEasedY, kCameraEasedScale;
float bgFade = 1.0f;

PImage textureImage;
PImage vignetteImage;

PFont titleFont;

Body[] bodies;

//================================================================
//MARK: Setup & Draw
//================================================================

public void setup() {
  //if you want to run in windowed mode, be sure to use P2D
  //size(1280, 960, P2D);

  //fullscreen:
  
  
  noCursor();
  frameRate(60);
  colorMode(HSB);

  //preferences
  loadPreferences();

  //vignette corners color setup
  kCornerTintLight = color(20, 255, 80); //rust red
  kCornerTintDark = color(190, 115, 80); //purple glow

  //title setup
  titleFont = createFont("Economica-Regular-OTF.otf", 40);
  textFont(titleFont);

  //background texture images
  vignetteImage = loadImage("clouds-no-vignette.png");
  textureImage = loadImage("ink-bg-texture-sand-white.png");

  //create creature bodies
  bodies = new Body[kBodyCount];
  for (int i=0; i<bodies.length; i++) {
    int bc = color(red(kBodyColorsHSB[i % kBodyColorsHSB.length]), 
      green(kBodyColorsHSB[i % kBodyColorsHSB.length]), 
      blue(kBodyColorsHSB[i % kBodyColorsHSB.length])
      );
    Body b = new Body(kSegmentCount, bc);
    bodies[i] = b;
  }

  //starting camera zoom
  kCameraEasedScale = 0.25f;
}

public void draw() {
  drawBackground();
  updateBodies();
  updateCamera();
  drawBodies();
}

//-------------------------------------------------
//drawBackground()
//fades solid color background and draws texture images
//-------------------------------------------------
public void drawBackground() {
  //background solid color
  if (millis() - kActiveStartTime > kPieceTimeInSeconds * 1000 &&
    kActive) {
    //main piece timing over, fading background and segments to black
    int c = color(hue(kBGColor), saturation(kBGColor), brightness(kBGColor) * bgFade);
    bgFade -= (1.0f / frameRate) / 20.0f;
    bgFade = max(0, bgFade);
    background(c);

    for (int i=0; i<bodies.length; i++) {
      bodies[i].scaleSegmentColorBrightness(bgFade);
    }
  } else {
    background(kBGColor);
  }

  //vignette
  beginShape();
  noStroke();
  texture(vignetteImage);
  if (bgFade < 1.0f) {
    colorMode(RGB);
    float r = red(kCornerTintLight) * bgFade + red(kCornerTintDark) * (1.0f - bgFade);
    float g = green(kCornerTintLight) * bgFade + green(kCornerTintDark) * (1.0f - bgFade);
    float b = blue(kCornerTintLight) * bgFade + blue(kCornerTintDark) * (1.0f - bgFade);
    tint(r, g, b);
    colorMode(HSB);
  } else {
    tint(kCornerTintLight);
  }
  textureWrap(REPEAT);
  textureMode(NORMAL);
  vertex(0, 0, 0, 0);
  vertex(width, 0, 1, 0);
  vertex(width, height, 1, 1);
  vertex(0, height, 0, 1);
  endShape();
  noTint();
}

//-------------------------------------------------
//updateBodies()
//calls update() on each body, also updates camera zoom
//-------------------------------------------------
public void updateBodies() {
  //update bodies
  for (int i=0; i<bodies.length; i++) {
    Body b = bodies[i];
    b.update();
  }
}

public void updateCamera() {
  //update camera
  //update / draw bodies
  float maxY = -1;
  float minY = MAX_FLOAT;
  for (int i=0; i<bodies.length; i++) {
    Body b = bodies[i];
    if (b.bodyStartPoint != null &&
      b.bodyStartPoint.y < minY) {
      minY = b.bodyStartPoint.y;
    }
    if (b.bodyEndPoint != null &&
      b.bodyEndPoint.y > maxY) {
      maxY = b.bodyEndPoint.y;
    }
  }

  //the smaller kCameraEasedScale, the more zoomed out
  float heightScale = abs(height / (maxY - minY));
  float widthScale = abs(width / (kBodySpacing * kBodyCount)) * 1.6f;
  float targetScale = min(heightScale, widthScale);
  targetScale = min(targetScale, kMaximumCameraZoom);
  float dScale = targetScale - kCameraEasedScale;
  float ease = (targetScale < kCameraEasedScale) ? 0.03f : 0.01f;
  kCameraEasedScale = kCameraEasedScale + (dScale * ease);

  //camera movement
  float targetY = (maxY + minY) / 2.0f;
  float dy = targetY - kCameraEasedY;
  if (frameCount < 20) {
    kCameraEasedY = kCameraEasedY + (dy);
  } else {
    kCameraEasedY = kCameraEasedY + (dy * 0.01f);
  }
}

//-------------------------------------------------
//drawBodies()
//zooms camera, draws bodies and text
//-------------------------------------------------
public void drawBodies() {
  pushMatrix();

  if (kRotate90) {
    translate(width / 2 + -kCameraEasedY * kCameraEasedScale, height / 2);
    rotate(PI / 2.0f);
  } else {
    translate(width / 2, height / 2 + kCameraEasedY * kCameraEasedScale);
  }
  scale(kCameraEasedScale, -kCameraEasedScale);

  beginShape();
  noStroke();
  texture(textureImage);
  tint(20, 255, 80, 15);
  textureWrap(REPEAT);
  textureMode(NORMAL);
  pushMatrix();
  rotate(PI / 2.0f);
  float sc = 2.5f;
  vertex(-width * sc, -height * sc, 0, 0);
  vertex(-width * sc, height * sc, 1, 0);
  vertex(width * sc, height * sc, 1, 1);
  vertex(width * sc, -height * sc, 0, 1);
  endShape();
  popMatrix();
  noTint();

  //draw bodies
  if (!kShowInstructions || kActive) {
    for (int i=0; i<bodies.length; i++) {
      Body b = bodies[i];
      float offset = (kBodyCount % 2 == 0) ? kBodySpacing / 2.0f + ((kBodyCount - 1) / 2) * kBodySpacing : ((kBodyCount - 1) / 2) * kBodySpacing;
      pushMatrix();
      translate(i * kBodySpacing - offset, 0);
      b.drawBody(null);
      popMatrix();
    }
  }

  //text
  if (!kActive) {
    textSize(40);
    textAlign(CENTER);
    noStroke();
    int c = color(hue(kBGColor), saturation(kBGColor), brightness(kBGColor) * 0.75f);
    fill(c);
    pushMatrix();
    rotate(PI);
    translate(0, height / 2.0f);
    scale(-1, 1);
    text("press spacebar to begin\npress 'i' for instructions", 0, 0);
    //text(kInstructionsString, 0, 0);
    popMatrix();
  }
  if (!kActive && kShowInstructions) {
    textSize(40);
    textAlign(CENTER);
    noStroke();
    int c = color(hue(kBGColor), saturation(kBGColor), brightness(kBGColor) * 0.65f);
    fill(c);
    pushMatrix();
    rotate(PI);
    scale(-1, 1);
    text(kInstructionsString, 0, 0);
    popMatrix();
  }

  popMatrix();
}

//================================================================
//MARK: Functions
//================================================================

public void loadPreferences() {
  String preferences[] = loadStrings("ink_well_preferences.txt");
  for (int i=0; i<preferences.length; i++) {
    String keyValue[] = split(preferences[i], '=');

    //kBodyCount
    if (keyValue[0].equals("creatureCount")) {
      kBodyCount = Integer.parseInt(keyValue[1]);
    }
    
    //kBodySpacing
    if (keyValue[0].equals("creatureSpacing")) {
      kBodySpacing = Float.parseFloat(keyValue[1]);
    }

    //kPieceTimeInSeconds
    if (keyValue[0].equals("duration")) {
      kPieceTimeInSeconds = Integer.parseInt(keyValue[1]);
    }

    //kRotate90
    if (keyValue[0].equals("screenRotate")) {
      kRotate90 = Boolean.parseBoolean(keyValue[1]);
    }
  }
}
//================================================================
//MARK: Class Body
//================================================================

class Body {

  //-------------------------------------------------
  //Public Properties
  //-------------------------------------------------
  
  float segmentCount;
  ArrayList<Segment> segments;

  float totalLength = 0.0f;
  float startPos, endPos;
  int waitTimer;

  float _cameraEasedY, _cameraEasedScale;
  PVector bodyStartPoint, bodyEndPoint; //should be readonly

  //-------------------------------------------------
  //Private Properties
  //-------------------------------------------------

  private Ramp startPosRamp, endPosRamp;
  private boolean startEndFlip = false;
  private Rune restRune;
  private PVector restPoint; 
  private boolean resting;
  private boolean waitingForRuneToHide;
  private int firstSegmentIndex, lastSegmentIndex;
  private int bodyColor;
  private float totalBodyLength;

  private PVector startRampValueDuration, endRampValueDuration;

  Body(int sCount, int _color) {
    segmentCount = sCount;

    endPosRamp = new Ramp();
    endPosRamp.sCurvePower = 6;

    startPosRamp = new Ramp();
    startPosRamp.sCurvePower = 6;

    bodyColor = _color;

    segments = new ArrayList<Segment>();

    int[] tendrilCount = new int[sCount];
    int[] rt = new int[sCount];
    for (int i=0; i<rt.length; i++) {
      rt[i] = i;
    }
    
    float randomPhase = random(TWO_PI);
    float freq = random(1.5f, 2.5f);
    for (int i=0; i<tendrilCount.length; i++) {
      float norm = PApplet.parseFloat(i) / tendrilCount.length;
      norm = abs(sin(norm * TWO_PI * freq + randomPhase));
      tendrilCount[i] = PApplet.parseInt(norm * norm * random(5, 20));
    }

    //growing plant top (will flip during draw) random
    PVector origin = new PVector(0, 0);
    PVector lastPoint = origin;
    bodyStartAnchorPoint = origin;
    for (int i=0; i<segmentCount; i++) {
      float x = lastPoint.x + (random(-kMaxRandomWalkWidth, kMaxRandomWalkWidth));
      x = max(min(x, width / 2.0f), -width / 2.0f);
      float y = lastPoint.y + 50;
      float l = random(kSegmentMaximumLength + kSegmentMinimumLength) + kSegmentMinimumLength;
      PVector rand = new PVector(x, y);
      Segment seg = new Segment(l, lastPoint, rand, tendrilCount[i]);
      segments.add(seg);
      totalLength += segments.get(i).len;
      lastPoint = segments.get(i).pointInSegment(1.0f);

      seg.setColor(bodyColor);
      seg.setColorBrightness(brightness(bodyColor) * 0.35f);

      //update previous cp1
      if (i > 0) {
        Segment s = segments.get(i);
        Segment prevS = segments.get(i-1);

        //get this segment's cp0, rotate it by theta, and set the previous to its opposite
        float cp0Len = sqrt(s.cp0.x*s.cp0.x + s.cp0.y*s.cp0.y);
        float theta = atan2(s.cp0.y, s.cp0.x) + PI;
        float dTheta = s.theta - prevS.theta;
        theta += dTheta;
        float newX = (cos(theta) * cp0Len) + prevS.len;
        float newY = (sin(theta) * cp0Len);
        prevS.cp1 = new PVector(newX, newY);
      }
    }
    totalBodyLength = totalLength;

    Segment last = segments.get(segments.size() - 1);
    float x = lastPoint.x + cos(last.theta) * last.len;
    float y = lastPoint.y + sin(last.theta) * last.len;
    bodyEndAnchorPoint = new PVector(x, y);

    float lengthSum = 0.0f;
    for (int i=0; i<segments.size(); i++) {
      Segment s = segments.get(i);
      s.bodyPosition = lengthSum;
      lengthSum += s.len;
      s.totalBodyLength = totalLength;
    }

    startPos = 0.0f;
    endPos = 0.0f;

    //we use {-1, -1} to indicate a 'null'
    startRampValueDuration = new PVector(-1, -1);
    endRampValueDuration = new PVector(-1, -1);

    //set up rest rune
    int runeColor = color(hue(bodyColor), saturation(bodyColor), max(brightness(bodyColor), 150));
    restRune = new Rune(200, runeColor);

    Segment firstSegment = segments.get(firstSegmentIndex);
    Segment lastSegment = segments.get(lastSegmentIndex);
    bodyStartPoint = firstSegment.startPoint();
    bodyEndPoint = lastSegment.endPoint();

    restPoint = new PVector(
      (bodyStartPoint.x + bodyEndPoint.x) / 2.0f, 
      (bodyStartPoint.y + bodyEndPoint.y) / 2.0f
      );
    setResting(true);
  }

  public void update() {
    if (kActive) {
      if (PApplet.parseInt(random(560)) == 0) {
        setResting(true);
      } else if (PApplet.parseInt(random(360)) == 0) {
        setResting(false);
      }
    }
    if (!resting) {
      if (millis() - waitTimer < 700) {
        waitingForRuneToHide = true;
        return;
      }

      if (waitingForRuneToHide) {
        startPosRamp.setValue(startPos);
        endPosRamp.setValue(endPos);
        waitingForRuneToHide = false;
      }

      boolean startAlmostFinished = startPosRamp.percentageComplete() > 0.8f;
      boolean endAlmostFinished = endPosRamp.percentageComplete() > 0.8f;
      if (startAlmostFinished &&
        endAlmostFinished) {
        if (startEndFlip) {
          startPosRamp.randomizeMode();
          startRampValueDuration.set(random(endPos), (random(500) + 2000) * kSpeedMultiplier);
        } else {
          endPosRamp.randomizeMode();
          float minLength = 100;
          float tep = min(totalLength, random(totalLength - startPos) + startPos + minLength);
          endRampValueDuration.set(tep, (random(500) + 2000) * kSpeedMultiplier);
        }
        startEndFlip = !startEndFlip;
      }
    }

    //check for ramps to start
    if (startRampValueDuration.y != -1) {
      startPosRamp.rampToValueOverDuration(startRampValueDuration.x, startRampValueDuration.y);
      startRampValueDuration.set(-1, -1);
    }
    if (endRampValueDuration.y != -1) {
      endPosRamp.rampToValueOverDuration(endRampValueDuration.x, endRampValueDuration.y);
      endRampValueDuration.set(-1, -1);
    }

    startPos = startPosRamp.update();
    endPos = endPosRamp.update();

    firstSegmentIndex = MAX_INT;
    lastSegmentIndex = -1;
    for (int i=0; i<segments.size(); i++) {
      Segment s = segments.get(i);
      boolean updated = s.update(startPos, endPos);
      if (updated) {
        if (i < firstSegmentIndex) {
          firstSegmentIndex = i;
        }
        if (i > lastSegmentIndex) {
          lastSegmentIndex = i;
        }
      }
    }

    if (lastSegmentIndex == -1 ||
      firstSegmentIndex == MAX_INT ||
      startPos > endPos) {
      //we're shrunk to zero, so...
      return;
    }

    Segment firstSegment = segments.get(firstSegmentIndex);
    Segment lastSegment = segments.get(lastSegmentIndex);
    bodyStartPoint = firstSegment.startPoint();
    bodyEndPoint = lastSegment.endPoint();
  }

  public void drawBody(PGraphics pg) {
    if (resting && startPos == endPos) {
      restRune.showAtPosition(restPoint);
    }
    if (restRune.isShowing()) {
      restRune.updateDraw(pg);
    } else {
      for (int i=0; i<segments.size(); i++) {
        Segment s = segments.get(i);
        s.drawSegment(pg);
      }
    }
  }

  public void setResting(boolean r) {
    if (r == resting) {
      return;
    }
    resting = r;
    if (!resting) {
      restRune.hide();
      waitTimer = millis();
    }
    if (resting) {
      float restPhase = (startPos + endPos) / 2.0f;
      startPosRamp.rampToValueOverDuration(restPhase, kRestRampTime);
      endPosRamp.rampToValueOverDuration(restPhase, kRestRampTime);

      PVector point = pointAtPhase(restPhase / totalBodyLength);
      if (point != null) {
        restPoint = pointAtPhase(restPhase / totalBodyLength);
        startEndFlip = false;
      } else {
        println("null restPoint restPhase "+restPhase+" tbl "+totalBodyLength);
      }
    }
  }

  //-------------------------------------------------
  //Functions
  //-------------------------------------------------

  public PVector pointAtPhase(float phase) {
    if (phase < 0.0f || phase > 1.0f) {
      return null;
    }

    float phaseLength = totalBodyLength * phase;

    float runningBodyLength = 0.0f;
    for (int i=0; i<segments.size(); i++) {
      Segment seg = segments.get(i);
      float startSegmentLength = runningBodyLength;
      runningBodyLength += seg.len;
      if (runningBodyLength > phaseLength) {
        //okay, this is the segment which has the phase
        float endSegmentLength = runningBodyLength;
        float segPhase = (phaseLength - startSegmentLength) / (endSegmentLength - startSegmentLength);
        PVector point = seg.pointInSegment(segPhase);
        return point;
      }
    }
    return null;
  }
  
  public void scaleSegmentColorBrightness(float b) {
      for (int i=0; i<segments.size(); i++) {
        Segment seg = segments.get(i);
        seg.scaleColorBrightness(b);
      } 
  }
}
//================================================================
//MARK: Globals
//================================================================

boolean kRotate90 = true;

int kBodyCount = 4;
float kBodySpacing = 250.0f;
int kSegmentCount = 50; //50

float kPieceTimeInSeconds = 1 * 60;

float kSegmentMinimumLength = 20;
float kSegmentMaximumLength = 60;

float kMaxRandomWalkWidth = 100;

float kBodyTendrilStartFadeLength = 300;
float kBodyTendrilEndFadeLength = 300;

float kMinimumTendrilCountPerSegment = 0;
float kMaximumTendrilCountPerSegment = 20;

float kMinimumTendrilLength = 50;
float kMaximumTendrilLength = 350;

float kSegmentStrokeWeight = 4;
float kTendrilStrokeWeight = 2;

float kSpeedMultiplier = 4.0f;
float kMaximumCameraZoom = 0.525f; //0.525 for 5

float kRestRampTime = 900.0f;
float kActiveStartTime;

//================================================================
//MARK: Interface
//================================================================

boolean kActive; //toggled to start
boolean kShowInstructions;
String kInstructionsString = "i. each player follows one ink creature\nii. when circle rune is showing, players must rest\niii. lines without tendrils are a single pitch\niv. tendrils are accents, timbre articulations, and wildly variable pitch\nv. dynamics should never go above piano";

//================================================================
//MARK: Colors
//================================================================

int kSegmentColor = color(50, 48, 30); //rgb
int kBGColor = color(215, 215, 175); //rgb

int kCornerTintLight;
int kCornerTintDark;

int[] kBodyColorsHSB = { //rgb here are representing hsb, set in setup()
  color(208.0f / 360 * 255, 0.01f * 255, 0.45f * 255), 
  color(42.0f / 360 * 255, 0.10f * 255, 0.65f * 255), 
  color(240.0f / 360 * 255, 0.0f * 255, 0.82f * 255), 
  color(55.0f / 360 * 255, 0.22f * 255, 0.64f * 255), 
  color(49.0f / 360 * 255, 0.43f * 255, 0.65f * 255)
};


//================================================================
//MARK: Enums
//================================================================

enum Side { 
  LEFTSIDE, RIGHTSIDE
}

//================================================================
//MARK: Bezier Functions
//================================================================

public void bezierTentacleWithPoints(float[] points, float thickness, PGraphics pg) {
  pg.beginShape();
  pg.beginShape();
  pg.vertex(points[0], points[1]);
  pg.bezierVertex(points[2], points[3], points[4], points[5], points[6], points[7]);
  pg.bezierVertex(points[4], points[5], points[2], points[3], points[0], points[1] + thickness);
  pg.endShape();
}

public void bezierWithPoints(float[] points) {
  bezier(
    points[0], 
    points[1], 
    points[2], 
    points[3], 
    points[4], 
    points[5], 
    points[6], 
    points[7]
    );
}

public void bezierWithPoints(float[] points, PGraphics pg) {
  pg.bezier(
    points[0], 
    points[1], 
    points[2], 
    points[3], 
    points[4], 
    points[5], 
    points[6], 
    points[7]
    );
}

public float[] reverseBezierPoints(float[] points) {
  float[] r = new float[points.length];
  r[0] = points[6];
  r[1] = points[7];
  r[2] = points[4];
  r[3] = points[5];
  r[4] = points[2];
  r[5] = points[3];
  r[6] = points[0];
  r[7] = points[1];
  return r;
}

public float[] splitBezier(float[] points, float t) {
  if (t == 1.0f) {
    return points;
  }
  float x1, y1, x2, y2, x3, y3, x4, y4 = 0.0f;
  x1 = points[0];
  y1 = points[1];
  x2 = points[2];
  y2 = points[3];
  x3 = points[4];
  y3 = points[5];
  x4 = points[6];
  y4 = points[7];

  float x12 = (x2-x1)*t+x1;
  float y12 = (y2-y1)*t+y1;

  float x23 = (x3-x2)*t+x2;
  float y23 = (y3-y2)*t+y2;

  float x34 = (x4-x3)*t+x3;
  float y34 = (y4-y3)*t+y3;

  float x123 = (x23-x12)*t+x12;
  float y123 = (y23-y12)*t+y12;

  float x234 = (x34-x23)*t+x23;
  float y234 = (y34-y23)*t+y23;

  float x1234 = (x234-x123)*t+x123;
  float y1234 = (y234-y123)*t+y123;

  float[] r = {
    x1, y1, x12, y12, x123, y123, x1234, y1234
  };
  return r;
}

public float[] splitBezier(float[] points, float t0, float t1) {
  if (t0 > t1) {
    return new float[8];
  }

  float[] r =  splitBezier(points, t1);
  float[] rev = reverseBezierPoints(r);
  float t = (t1 - t0) / t1;
  r = splitBezier(rev, t);
  return r;
}
//================================================================
//MARK: Interface
//================================================================

public void keyPressed() {
  if (keyCode == 32) { //spacebar
    kActive = true;
    kActiveStartTime = millis();
  }
  if (key == 's') {
    //screenshot saving
    saveFrame("screenshots/ink-well-screenshot-"+year()+""+month()+""+day()+""+hour()+""+minute()+""+second()+".png");
  }
  if (key == 'i') {
    kShowInstructions = !kShowInstructions;
  }
}
//================================================================
//MARK: Ramp
//================================================================

enum RampMode {
  RAMP_MODE_LINEAR, 
    RAMP_MODE_EASE_IN, 
    RAMP_MODE_EASE_OUT, 
    RAMP_MODE_S_CURVE
}

class Ramp {

  //================================================================
  //MARK: Public Properties
  //================================================================

  float[] valueArray;
  float value = 0.0f;
  RampMode mode = RampMode.RAMP_MODE_LINEAR;
  float sCurvePower = 1;

  //================================================================
  //MARK: Private Properties
  //================================================================

  private float start;
  private float end;

  private float[] startArray;
  private float[] endArray;

  private float startTime;
  private float duration;
  private boolean running;


  //================================================================
  //MARK: Readonly Properties
  //================================================================

  public float percentageComplete() {
    if (duration == 0) {
      return 1.0f;
    }
    float m = millis();
    float p = (m - startTime) / duration;
    //println("percCom m st d "+p+" "+m+" "+startTime+" "+duration);
    return p;
  }

  //================================================================
  //MARK: Initialization
  //================================================================

  Ramp() {
    start = 0.0f;
    end = 0.0f;
    duration = 0;
    value = 0.0f;
    startTime = 0;

    running = false;
  }

  //================================================================
  //MARK: Value Changing Functions
  //================================================================

  public void rampToValueOverDuration(float v, float d) {
    if (d == 0) {
      setValue(v);
    } else {
      start = value;
      end = v;
      duration = d;
      start();
    }
  }

  public void setValue(float v) {
    start = v;
    end = v;
    value = v;
    duration = 0;
    start();
  }

  public void setValues(float[] v) {
    startArray = new float[v.length];
    for (int i=0; i<v.length; i++) {
      startArray[i] = v[i];
    }
    endArray = new float[v.length];
    for (int i=0; i<v.length; i++) {
      endArray[i] = v[i];
    }
    valueArray = new float[v.length];
    for (int i=0; i<v.length; i++) {
      valueArray[i] = v[i];
    }
    duration = 0;
    start();
  }

  public void rampToValuesOverDuration(float[] v, float d) {
    if (d == 0) {
      setValues(v);
    } else {
      startArray = new float[v.length];
      for (int i=0; i<v.length; i++) {
        if (valueArray != null &&
          valueArray.length > i) {
          startArray[i] = valueArray[i];
        } else {
          startArray[i] = 0;
        }
      }
      endArray = new float[v.length];
      for (int i=0; i<v.length; i++) {
        endArray[i] = v[i];
      }
      valueArray = new float[v.length];
      for (int i=0; i<v.length; i++) {
        valueArray[i] = startArray[i];
      }
      duration = d;
      start();
    }
  }

  //================================================================
  //MARK: Start / Stop / Update
  //================================================================

  public void start() {
    startTime = millis();
    running = true;
  }

  public void stop() {
    running = false;
  }

  public float update() {
    if (isFinished()) {
      running = false;
      value = end;
      return end;
    }

    if (!running) {
      start();
    }

    //scale value
    value = (end - start) * normalizedPhase() + start;

    if (valueArray != null &&
      valueArray.length != 0) {
      for (int i=0; i<valueArray.length; i++) {
        valueArray[i] = (endArray[i] - startArray[i]) * normalizedPhase() + startArray[i];
      }
    }

    return value;
  }

  //================================================================
  //MARK: Normalized Value Calculation
  //================================================================

  public float normalizedPhase() {
    if (duration == 0) {
      return 1.0f;
    }
    float n = (millis() - startTime) / duration;
    switch (mode) {
    case RAMP_MODE_EASE_IN:
      {
        n *= n;
      }
      break;
    case RAMP_MODE_EASE_OUT:
      {
        n = 1.0f - (1.0f - n) * (1.0f - n);
      }
      break;
    case RAMP_MODE_S_CURVE:
      {
        float dn = 1.0f - pow(1.0f - n, sCurvePower);
        n = 0.5f - (cos(dn * PI) * 0.5f);
      }
      break;
    case RAMP_MODE_LINEAR:
    default:
      break;
    }
    n = max(0.0f, min(n, 1.0f));
    return n;
  }

  //================================================================
  //MARK: Readonly Getters
  //================================================================

  public boolean isRunning() {
    return running;
  }

  public boolean isFinished() {
    return millis() >= startTime + duration;
  }

  public int intValue() {
    return PApplet.parseInt(value);
  }

  public float startValue() {
    return start;
  }

  public float endValue() {
    return end;
  }

  //================================================================
  //MARK: Class Functions
  //================================================================

  public void randomizeMode() {
    int i = PApplet.parseInt(random(4));
    switch (i) {
    case 0: 
      mode = RampMode.RAMP_MODE_LINEAR; 
      break;
    case 1: 
      mode = RampMode.RAMP_MODE_EASE_IN;
      break;
    case 2: 
      mode = RampMode.RAMP_MODE_EASE_OUT; 
      break;
    case 3: 
      mode = RampMode.RAMP_MODE_S_CURVE;
      break;
    default:
    }
  }
}
//================================================================
//MARK: Rune
//================================================================

class Rune {

  //-------------------------------------------------
  //Private Properties
  //-------------------------------------------------

  private float maxDiameter;
  private PVector center;
  private int runeColor;

  private Ramp scaleRamp, rotationRamp;

  private PGraphics pg;
  private boolean showing;

  //-------------------------------------------------
  //Readonly Properties
  //-------------------------------------------------

  public boolean isShowing() {
    return showing;
  }

  //-------------------------------------------------
  //Initializer
  //-------------------------------------------------
  Rune(float diameter, int r) {
    maxDiameter = diameter;
    center = new PVector(0, 0);
    scaleRamp = new Ramp();
    scaleRamp.mode = RampMode.RAMP_MODE_S_CURVE;
    rotationRamp = new Ramp();
    rotationRamp.mode = RampMode.RAMP_MODE_S_CURVE;
    runeColor = r;

    showing = false;

    pg = createGraphics(PApplet.parseInt(maxDiameter) + 10, PApplet.parseInt(maxDiameter) + 10, P2D); //10 to allow for strokeWeight / other edges
    drawGraphics();
  }

  public void updateDraw(PGraphics p) {
    float scale = scaleRamp.update();
    if (scale == 0.0f && scaleRamp.isFinished()) {
      showing = false;
      return;
    }
    float rotation = rotationRamp.update() * PI;
    float generalRotation = PApplet.parseFloat(frameCount) * 0.002f;

    if (p != null) {
      p.pushMatrix();
      p.translate(center.x, center.y);
      p.rotate(rotation + generalRotation);
      p.scale(scale);
      p.image(pg, -pg.width / 2, -pg.height / 2);
      p.popMatrix();
    } else {
      pushMatrix();
      translate(center.x, center.y);
      rotate(rotation + generalRotation);
      scale(scale);
      image(pg, -pg.width / 2, -pg.height / 2);
      popMatrix();
    }
  }

  public void showAtPosition(PVector p) {
    if (showing) {
      return;
    }
    center = p;
    scaleRamp.rampToValueOverDuration(1.0f, 1200);
    rotationRamp.rampToValueOverDuration(1.0f, 1600);
    showing = true;
  }

  public void hide() {
    if (!showing) {
      return;
    }
    scaleRamp.rampToValueOverDuration(0.0f, 600);
    rotationRamp.rampToValueOverDuration(2.0f, 800);
  }

  private void drawGraphics() {
    pg.beginDraw();
    pg.pushMatrix();
    pg.translate(pg.width / 2, pg.height / 2);
    pg.stroke(runeColor);
    pg.noFill();
    //pg.fill(runeColor);
    pg.strokeWeight(2);

    float theta = random(PI * 2.0f);

    ArrayList<PVector> centers = new ArrayList<PVector>();

    //pick a starting edge
    float radius = 5.0f;
    for (int n=0; n<3; n++) {
      PVector start = new PVector(cos(theta) * (100.0f - radius), sin(theta) * (100.0f - radius));
      centers.add(start);
      pg.ellipse(start.x, start.y, radius * 2, radius * 2);
      for (int i=0; i<100; i++) {
        //get last point
        PVector last = centers.get(centers.size() - 1);
        //get a random position within the circle
        PVector rand = new PVector(
          cos(random(PI * 2.0f)) * 100, 
          sin(random(PI * 2.0f)) * 100
          );
        PVector next = new PVector(
          last.x + cos(theta) * radius * 2.0f, 
          last.y + sin(theta) * radius * 2.0f
          );
        //get the angle to that point
        theta = atan2(rand.y - last.y, rand.x - last.x);
        //hit test with others
        boolean hitOther = false;
        for (int j=0; j<centers.size(); j++) {
          PVector c = centers.get(j);
          float dc = next.dist(c);
          if (dc < radius * 2) {
            hitOther = true;
          }
        }
        if (next.dist(new PVector(0, 0)) < (100.0f - radius) &&
          !hitOther) {
          pg.stroke(runeColor, random(256));
          pg.noFill();
          pg.fill(runeColor, random(256));
          pg.ellipse(next.x, next.y, radius * 2, radius * 2);
          centers.add(next);
        }
      }
    }

    pg.stroke(runeColor);
    pg.noFill();
    pg.ellipse(0, 0, 200, 200);

    int circleCount = 100;
    float dTheta = (PI * 2.0f) / PApplet.parseFloat(circleCount);
    for (int i=0; i<circleCount; i++) {
      if (random(3) > 1.0f) {
        pg.noStroke();
        pg.stroke(runeColor);
        pg.fill(0);
        PVector p = new PVector(cos(dTheta * i) * 100.0f, sin(dTheta * i) * 100);
        pg.ellipse(p.x, p.y, 6, 6);
      }
    }
    pg.popMatrix();
    pg.endDraw();
  }
}
//================================================================
//MARK: Class Segment
//================================================================

class Segment {

  //-------------------------------------------------
  //Public Properties
  //-------------------------------------------------

  float len;
  float theta;
  float bodyPosition, totalBodyLength;
  PVector anchor;

  float segmentStartPhase, segmentEndPhase;
  PVector cp0, cp1;

  boolean drawTendrils = true;

  //-------------------------------------------------
  //Private Properties
  //-------------------------------------------------

  private float _startBodyPosition, _endBodyPosition;
  private Tendril[] tendrils;
  private boolean tendrilsRequireUpdate; //we recalc tendril positions until we get a full phase draw
  private int segmentColor; //use setColor
  private float tendrilSinFrequency = random(2.2f, 3.3f);
  private int originalColor;

  //-------------------------------------------------
  //Lifecycle
  //-------------------------------------------------

  Segment(float _length, PVector _anchor, PVector _theta, int _tendrilCount) {
    len = _length;
    theta = atan2(_theta.y - _anchor.y, _theta.x - _anchor.x); //note: atan2(y, x);
    anchor = _anchor;

    segmentStartPhase = 0.0f;
    segmentEndPhase = 0.0f;

    float scx0 = random(len * 0.9f) + len * 0.1f;
    float scy0 = 10;
    float scx1 = 0;
    float scy1 = 0;

    cp0 = new PVector(scx0, scy0);
    cp1 = new PVector(scx1, scy1);

    //tendrils = new ArrayList<Tendril>();
    tendrils = new Tendril[_tendrilCount];
    //int tendrilCount = int(random(kMinimumTendrilCountPerSegment, kMaximumTendrilCountPerSegment));
    float scale = 1.0f - (_tendrilCount / 20.0f);
    for (int i=0; i<_tendrilCount; i++) {
      Side side = (PApplet.parseInt(random(2)) == 0) ? Side.LEFTSIDE : Side.RIGHTSIDE;
      tendrils[i] = new Tendril(random(kMinimumTendrilLength, kMaximumTendrilLength) * scale, random(1), side);
      //tendrils.add(tendril);
    }

    tendrilsRequireUpdate = true;
  }

  //-------------------------------------------------
  //Update & Draw
  //-------------------------------------------------

  /**
   returns TRUE if update is necessary, FALSE if not
   */
  public boolean update(float startBodyPosition, float endBodyPosition) {
    _startBodyPosition = startBodyPosition;
    _endBodyPosition = endBodyPosition;

    //we know our anchorPosition and len, so we can calculate how much of this segment to draw
    //we can also fade in knowing the start and end body positions

    if (startBodyPosition > bodyPosition + len ||
      endBodyPosition < bodyPosition) {
      return false;
    }

    segmentEndPhase = min((endBodyPosition - bodyPosition) / len, 1.0f);
    segmentStartPhase = min((max(startBodyPosition, bodyPosition) - bodyPosition) / len, 1.0f);

    return true;
  }

  /**
   returns TRUE if update is necessary, FALSE if not
   */
  public boolean drawSegment(PGraphics pg) {
    if (_startBodyPosition > bodyPosition + len ||
      _endBodyPosition < bodyPosition) {
      return false;
    }

    if (pg != null) {
      pg.pushMatrix();

      //translate to anchorpoint and rotate
      pg.translate(anchor.x, anchor.y);
      pg.rotate(theta);
    } else {
      pushMatrix();

      //translate to anchorpoint and rotate
      translate(anchor.x, anchor.y);
      rotate(theta);
    }

    if (drawTendrils) {
      for (int i=0; i<tendrils.length; i++) {
        Tendril tendril = tendrils[i];
        if (segmentStartPhase < tendril.phase &&
          segmentEndPhase > tendril.phase) {
          float tendrilBodyPosition = bodyPosition + (tendril.phase * len);
          float endScale = max(min((_endBodyPosition - tendrilBodyPosition) / kBodyTendrilEndFadeLength, 1.0f), 0.0f);
          float startScale = max(min((tendrilBodyPosition - _startBodyPosition) / kBodyTendrilStartFadeLength, 1.0f), 0.0f);

          float tendrilScale = startScale * endScale;
          float side = (tendril.side == Side.LEFTSIDE) ? 1.0f : -1.0f;
          //float side = 1.0;

          //float frequency = 3.0;
          float tendrilAngle = (tendrilBodyPosition / (totalBodyLength) * PI * tendrilSinFrequency) + PI + PI * 0.5f;
          //float tendrilAngle = tendril.phase * TWO_PI;

          if (pg != null) {
            pg.pushMatrix();
          } else {
            pushMatrix();
          }
          if (tendrilsRequireUpdate) {
            updateTendrilPositions();
            if (segmentStartPhase == 0.0f &&
              segmentEndPhase == 1.0f) {
              tendrilsRequireUpdate = false;
            }
          }
          if (pg != null) {
            pg.translate(tendril.anchor.x, tendril.anchor.y);
            pg.rotate(-theta + (tendrilAngle * side));
            pg.stroke(tendril.col);
            pg.noFill();
            pg.strokeWeight(kTendrilStrokeWeight);
          } else {
            translate(tendril.anchor.x, tendril.anchor.y);
            rotate(-theta + (tendrilAngle * side));
            stroke(tendril.col);
            noFill();
            strokeWeight(kTendrilStrokeWeight);
          }
          //line(0, 0, tendril.len * tendrilScale * side, 0);

          //tendrilScale = 1.0;
          float scaler = tendrilScale * side;
          float[] tbp = {0, 
            0, 
            tendril.cp0.x * scaler, 
            tendril.cp0.y, 
            tendril.cp1.x * scaler, 
            tendril.cp1.y, 
            tendril.len * scaler, 
            0}; 
          tbp = splitBezier(tbp, tendrilScale);
          if (pg != null) {
            bezierWithPoints(tbp, pg);
          } else {
            bezierWithPoints(tbp);
          }

          if (pg != null) {
            pg.popMatrix();
          } else {
            popMatrix();
          }
        }
      }
    }

    //from here out we consider the body of the segment to be horizontal,
    //tendrils will branch out from top (LEFTSIDE) and bottom (RIGHTSIDE)
    if (pg != null) {
      pg.stroke(segmentColor);
      pg.noFill();
      pg.strokeWeight(kSegmentStrokeWeight);
    } else {
      stroke(segmentColor);
      noFill();
      strokeWeight(kSegmentStrokeWeight);
    }

    float x0 = len * segmentStartPhase;
    float x1 = len * segmentEndPhase;
    float[] bp = {x0, 0, cp0.x, cp0.y, cp1.x, cp1.y, x1, 0};
    //if we're only drawing a portion of the segment...
    if (segmentStartPhase != 0.0f ||
      segmentEndPhase != 1.0f) {
      bp = splitBezier(bp, segmentStartPhase, segmentEndPhase);
    }
    if (pg != null) {
      bezierWithPoints(bp, pg);
    } else {
      bezierWithPoints(bp);
    }

    if (pg != null) {
      pg.popMatrix();
    } else {
      popMatrix();
    }
    return true;
  }

  //-------------------------------------------------
  //Functions
  //-------------------------------------------------

  public void updateTendrilPositions() {
    float x0 = len * segmentStartPhase;
    float x1 = len * segmentEndPhase;   
    float[] bp = {x0, 0, cp0.x, cp0.y, cp1.x, cp1.y, x1, 0};
    for (int i=0; i<tendrils.length; i++) {
      Tendril tendril = tendrils[i];
      float x = bezierPoint(bp[0], bp[2], bp[4], bp[6], tendril.phase);
      float y = bezierPoint(bp[1], bp[3], bp[5], bp[7], tendril.phase);
      tendril.anchor = new PVector(x, y);
    }
  }

  public PVector startPoint() {
    return pointInSegment(segmentStartPhase);
  }

  public PVector endPoint() {
    return pointInSegment(segmentEndPhase);
  }

  public PVector pointInSegment(float phase) {
    float dx = cos(theta) * (len * phase);
    float dy = sin(theta) * (len * phase);
    return new PVector(anchor.x + dx, anchor.y + dy);
  }

  public void setColor(int c) {
    segmentColor = c;
    originalColor = c;
    for (int i=0; i<tendrils.length; i++) {
      tendrils[i].setColor(segmentColor);
    }
  }

  public void setColorSaturation(float s) {
    for (int i=0; i<tendrils.length; i++) {
      tendrils[i].setColorSaturation(s);
    }
  }

  public void scaleColorBrightness(float b) {
    segmentColor = color(hue(originalColor), saturation(originalColor), brightness(originalColor) * b);
  }

  public void setColorBrightness(float b) {
    segmentColor = color(hue(segmentColor), saturation(segmentColor), b);
    originalColor = segmentColor;
  }
}
//================================================================
//MARK: Class Tendril
//================================================================

class Tendril {

  //-------------------------------------------------
  //Properties
  //-------------------------------------------------
  float len, phase;
  Side side;

  PVector anchor;
  PVector cp0, cp1;
  int col;

  private int originalColor;

  Tendril(float _length, float _phase, Side _side) {
    len = _length;
    phase = _phase;
    side = _side;

    float cpx0 = random(len * 0.5f) + len * 0.1f;
    float cpy0 = random(len * 0.4f) - len * 0.2f;
    cp0 = new PVector(cpx0, cpy0);

    float cpx1 = random(len * 0.3f) + len * 0.4f;
    float cpy1 = random(len * 0.4f) - len * 0.2f;
    cp1 = new PVector(cpx1, cpy1);

    anchor = new PVector(-1, -1);
  }

  public void setColor(int c) {
    float hue, sat, bri;
    hue = hue(c);
    sat = saturation(c);
    bri = PApplet.parseInt(random(brightness(c) - 55) + 55);
    col = color(hue, sat, bri);
    originalColor = col;
  }

  public void setColorSaturation(float s) {
    col = color(hue(originalColor), s, brightness(originalColor));
  }

  public void setColorBrightness(float b) {
    col = color(hue(originalColor), saturation(originalColor), b);
  }
}
  public void settings() {  fullScreen(P2D);  pixelDensity(displayDensity()); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "--present", "--window-color=#000000", "--hide-stop", "ink_well" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
